/*
 * pstack - 饤֥ʬ (¾ΥץꥱȤ߹)
 *
 * (C)Copyright 2009, Yashiro Takeshi <yashiromann@nifty.com>
 */

#include <basic.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <tcode.h>
#include <tstring.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <btron/errcode.h>
#include <btron/tkcall.h>
#include <libelf/libelf.h>
#include <libelf/gelf.h>
#include "pstack.h"

#if defined(USE_DEMANGLE) && USE_DEMANGLE
extern char* cpp_demangle(const char* mangled, char* output,
                          size_t* len, int* status);
#endif

//#define NDEBUG

#define MAX_PC 128
#define PSTKDRV_DEVNM "pstkdrv"

#ifdef _PCAT_
#define PCEQ_PREFIX "EIP ="
#else
#define PCEQ_PREFIX " PC =" 
#endif


static PstkSymbolTable* init_symtab()
{
    PstkSymbolTable* tbl = 
        (PstkSymbolTable*)malloc(sizeof(PstkSymbolTable));
    if (tbl == NULL) {
        printf("WARNING: Out of memory\n");
        return NULL;
    }

    tbl->n    = 0;
    tbl->nmax = 256;
    tbl->defs = (PstkSymbolDef*)malloc(tbl->nmax * sizeof(PstkSymbolDef));
    if (tbl->defs == NULL) {
        printf("WARNING: Out of memory\n");
        free(tbl);
        return NULL;
    }

    return tbl;
}


static W add_symtab(PstkSymbolTable* tbl, UW addr, const char* name)
{
    int nmax;
    PstkSymbolDef* p;

    if (name == NULL) {
        name = "";
    }

    nmax = tbl->nmax;
    while (tbl->n >= nmax) {
        nmax *= 2;
    }

    p = (PstkSymbolDef*)realloc(tbl->defs, nmax * sizeof(PstkSymbolDef));
    if (p == NULL) {
        printf("WARNING: Out of memory\n");
        return ER_NOMEM;
    }
    tbl->defs = p;
    tbl->nmax = nmax;

    tbl->defs[tbl->n].addr = addr;
    tbl->defs[tbl->n].name = strdup(name);
    tbl->n++;
    assert(tbl->n <= tbl->nmax);

    return 0;
}



static PstkSymbolTable* load_symtab_exe(const char* fname)
{
    Elf* elf;
    GElf_Shdr shdr;
    Elf_Scn* scn;
    Elf_Data* data;
    GElf_Sym sym;
    int type;
    PstkSymbolTable* tbl = NULL;
    int fd, i;
    W err;
#if defined(USE_DEMANGLE) && USE_DEMANGLE
    int status;
#endif

    elf_version(EV_CURRENT);

    fd = open(fname, O_RDONLY);
    if (fd < 0) {
        printf("WARNING: Unable to open file.\n");
        goto err_ret0;
    }

    elf = elf_begin(fd, ELF_C_READ, NULL);
    if (elf == NULL) {
        goto err_ret1;
    }

    scn = NULL;
    while ((scn = elf_nextscn(elf, scn)) != NULL) {
        gelf_getshdr(scn, &shdr);
        if (shdr.sh_type == SHT_SYMTAB) {
            break;
        }
    }
    if (scn == NULL) {
        goto err_ret2;
    }

    tbl = init_symtab();
    if (tbl == NULL) {
        goto err_ret2;
    }

    data = NULL;
    while ((data = elf_getdata(scn, data)) != NULL) {
        int count = shdr.sh_size / shdr.sh_entsize;
        for (i = 0; i < count; i++) {
            gelf_getsym(data, i, &sym);

            type = GELF_ST_TYPE(sym.st_info);
            /* FIXME: ʲξϤʤꤤ */
            if (type != STT_FILE && type != STT_SECTION) {
                char* p = elf_strptr(elf, shdr.sh_link, sym.st_name);
#if defined(USE_DEMANGLE) && USE_DEMANGLE
                if (p != NULL) {
                    char* q = cpp_demangle(p, 0, 0, &status);
                    if (q != NULL) {
                        p = q;
                    }
                }
#endif
                err = add_symtab(tbl, (UW)(sym.st_value), p);
                if (err < 0) {
                    printf("WARNING: Out of memory.\n");
                    goto err_ret3;
                }
            }
        }
    }

    elf_end(elf);
    close(fd);
    return tbl;

err_ret3:
    free_symtab(tbl);
err_ret2:
    elf_end(elf);
err_ret1:
    close(fd);
err_ret0:
    return NULL;
}


static int char_to_hex(int c)
{
    if (c >= '0' && c <= '9') {
        return c - '0';
    }
    if (c >= 'a' && c <= 'f') {
        return c - 'a' + 10;
    }
    if (c >= 'A' && c <= 'F') {
        return c - 'A' + 10;
    }
    assert(FALSE);
    return -1;
}


static PstkSymbolTable* load_symtab_map(const char* fname)
{
    FILE* fp;
    char line[4096];
    PstkSymbolTable* tbl = NULL;
    W err;
    UW addr;
    BOOL cont = FALSE;
    int i;

    fp = fopen(fname, "r");
    if (fp == NULL) {
        goto err_ret0;
    }

    tbl = init_symtab();
    if (tbl == NULL) {
        goto err_ret1;
    }

    while (fgets(line, sizeof(line), fp) != NULL) {
        char* p = strchr(line, '\n');
        if (p != NULL) {
            *p = '\0';
            if (cont) {
                /* ĹԤ³ʤΤǡparseʤǼΤƤ */
                cont = FALSE;
                continue;
            }
        }
        else {
            printf("WARNING: Ignored a line too long.\n");
            cont = TRUE;
            continue;
        }

        /* FIXME: ʲparseϤʤꤤø⡼ */
        if (strlen(line) <= 10 
            || line[8] != ' ' || line[10] != ' '
            || !isalpha(line[9])) {
            continue;
        }
        if (line[9] == 'U' || line[9] == 'u') {
            continue;
        }
        addr = 0;
        for (i = 0; i < 8; i++) {
            if (!isxdigit(line[i])) {
                break;
            }
            addr *= 0x10;
            addr += char_to_hex(line[i]);
        }
        if (i != 8) {
            continue;
        }
        err = add_symtab(tbl, addr, line + 11);
        if (err < 0) {
            printf("WARNING: Out of memory.\n");
            goto err_ret2;
        }
    }

    fclose(fp);
    return tbl;

err_ret2:
    free_symtab(tbl);
err_ret1:
    fclose(fp);
err_ret0:
    return NULL;
}


static int compare_symtab(const void* a, const void* b)
{
    W a_, b_;

    a_ = ((PstkSymbolDef*)a)->addr;
    b_ = ((PstkSymbolDef*)b)->addr;
    if (a_ > b_) {
        return 1;
    }
    else if (a_ < b_) {
        return -1;
    }
    return 0;
}


static inline void sort_symtab(PstkSymbolTable* tbl)
{
    qsort(tbl->defs, tbl->n, sizeof(PstkSymbolDef), compare_symtab);
}


PstkSymbolTable* load_symtab(const char* fname)
{
    PstkSymbolTable* tbl = load_symtab_exe(fname);
    if (tbl == NULL) {
        tbl = load_symtab_map(fname);
        if (tbl == NULL) {
            printf("WARNING: Invalid exec/map file format.\n");
            goto err_ret0;
        }
    }

    sort_symtab(tbl);
    return tbl;

err_ret0:
    return NULL;
}


void free_symtab(PstkSymbolTable* tbl)
{
    if (tbl != NULL) {
        if (tbl->defs != NULL) {
            int i;
            for (i = 0; i < tbl->n; i++) {
                free(tbl->defs[i].name);
            }
            free(tbl->defs);
        }
        free(tbl);
    }
    return;
}


const PstkSymbolDef* resolve_symbol(PstkSymbolTable* tbl, UW addr)
{
    int ibegin, iend, pivot;

    if (tbl == NULL) {
        return NULL;
    }

    ibegin = 0;
    iend = tbl->n;
    while (iend - ibegin > 1) {
        pivot = (ibegin + iend) / 2;
        if (addr < tbl->defs[pivot].addr) {
            iend = pivot;
        }
        else {
            ibegin = pivot;
        }
    }
    return &(tbl->defs[ibegin]);
}


W print_stacktrace(W tid, PstkSymbolTable* tbl)
{
    W err;
    VP buf[MAX_PC];
    ID dd = -1;
    INT asize;
    int i;

    err = tk_opn_dev(PSTKDRV_DEVNM, TD_READ);
    if (err < 0) {
        printf("ERROR: Driver not loaded.\n");
        printf("       Try \"lodspg pstkdrv\" unless already done.\n");
        goto err_ret0;
    }
    dd = err;

    err = tk_srea_dev(dd, tid, buf, sizeof(buf), &asize);
    if (err < 0 || asize <= 0) {
        printf("ERROR: Failed to get task information from driver.\n");
        printf("       Probably tid=%d is invalid.\n", tid);
        goto err_ret1;
    }

    tk_cls_dev(dd, 0);
    dd = -1;

    printf("*** TID = %d:\n", tid);
    for (i = 0; i < asize/sizeof(VP); i++) {
        const PstkSymbolDef* def = resolve_symbol(tbl, (UW)(buf[i]));
        const char* prefix = (i == 0) ? PCEQ_PREFIX : "  <--";

        if (def != NULL) {
            printf("%s %08x : %s+%u\n", prefix, (UW)(buf[i]), 
                   def->name, (UW)(buf[i]) - (UW)(def->addr));
        }
        else {
            printf("%s %08x\n", prefix, (UW)(buf[i]));
        }
    }
    printf("\n");

    return 0;

err_ret1:
    if (dd > 0) {
        tk_cls_dev(dd, 0);
    }

err_ret0:    
    printf("\n");
    return err;
}


